// This work is published under the GNU Public License (GPL)
// see file COPYING for details.

package org.gnugk;

import java.io.*;
import java.net.Socket;
import java.util.Vector;

/**
 * Communication thread for GnuGk status port.
 *
 * @author Jan Willamowius <jan@willamowius.de>
 * @version 1.3
 */

public class GkClient implements Runnable {
    private CallRouter router;
	private String host;
	private int port;
	
	// socket io to communicate with the gatekeeper
	private Socket socket;
	private InputStream istream;
	private OutputStream ostream;
    private volatile boolean stopped = false;
    private static final long WAIT_SEC = 5; // wait 5 seconds before reconnecting to GnuGk

    /**
     * Create a new thread; you must call init() before it can be run().
     */
    public GkClient() {
	}

    /**
     * Initialize with connection parameters to GnuGk.
     * @param _store    where to store agent states
     * @param _host     GnuGk host
     * @param _port     GnuGk status port
     */
	public void init(CallRouter _store, String _host, int _port) {
        router = _store;
		host = _host;
		port = _port;
	}

    /**
     * Stop this thread.
     */
	public void stop() {
        stopped = true;
        close();
    }

    /**
     * close connection to GnuGk.
     */
    private void close() {
		try {
			// cleanup
			if (istream != null)
				istream.close();
			istream = null;
			if (ostream != null)
				ostream.close();
			ostream = null;
			if (socket != null)
				socket.close();
			socket = null;
		}
		catch(IOException e) { return; };
	};

    /**
     * Run this thread: Connect to GnuGk and process the events.
     */
	public synchronized void run() {
		if (istream != null) {
			return;
		}
        while (!stopped) {
            try {
                Logger.log("Connecting to gatekeeper at " + host + ":" + port);
                socket = new Socket(host, port);
                ostream = socket.getOutputStream();
                istream = socket.getInputStream();
                BufferedReader in = new BufferedReader(new InputStreamReader(istream));

                router.clearAllAliasStates();
                refresh();	// get initial gk status
                Logger.log("Connection established.");
                String inputLine;
                while ((inputLine = in.readLine()) != null) {
                    // append the event to log
                    Logger.debug(inputLine);
                    String [] msgField = splitFields(inputLine);

                    // update endpoint buttons
                    if (inputLine.startsWith("RCF") && (msgField.length >= 5)) {
                        String epid = msgField[4];
                        // extract _list_ of aliases and add all!
                        String [] alias = splitAliases(msgField[2]);
                        for (int i=0; i < alias.length; i++) {
                            router.addAlias(alias[i], epid);
                        }
                    }
                    else if (inputLine.startsWith("UCF") && (msgField.length >= 3)) {
                        String epid = msgField[2];      // remove _all_ aliases for this epid
                        router.removeAllAliases(epid);
                    }
                    else if (inputLine.startsWith("ACF") && (msgField.length >= 3)) {
                        String crv = msgField[3];       // remember CRV for DCF
                        String [] alias = splitAliases(msgField[4]);    // destinationInfo
                        for (int i=0; i < alias.length; i++) {
                            router.setAliasTalking(alias[i], crv);
                        }
                        alias = splitAliases(msgField[5]);    // srcInfo
                        for (int i=0; i < alias.length; i++) {
                            router.setAliasTalking(alias[i], crv);
                        }
                    }
                    else if (inputLine.startsWith("DCF") && (msgField.length >= 3)) {
                        String epid = msgField[2];
                        String crv = msgField[3];     // identify by epid + CRV
                        router.setAliasAvailable(epid, crv);
                    }
                    else if (inputLine.startsWith("RouteRequest") && (msgField.length >= 6)) {
                        String callingEpId = msgField[2];
                        String callref = msgField[3];
                        String queue = msgField[4];
                        router.routeCall(queue, callingEpId, callref, msgField[5], msgField[1] );
                    }
                }
            }
            catch(IOException e) {
                // communication with GnuGk failed
                Logger.log("Error: " + e);
                Logger.log("Lost contact to GnuGk, will try reconnect in " + WAIT_SEC + " seconds");
                close();     // close connection
                try {
                    Thread.sleep(WAIT_SEC * 1000);     // sleep a little before trying again
                } catch (InterruptedException ie) {
                    // ignore, will reconnect a little earlier, but that doesn't matter
                }
            };
        }
		Logger.log("Disconnecting from gatekeeper");
	}

    /**
     * Split GnuGk events into files (separated by pipe char).
     * @param aString   string to split
     * @return          array of fields
     */
    private static String [] splitFields(String aString) {
        // remove whitespave
        aString = aString.trim();
        // remove semicolon
        if (aString.endsWith(";")) {
            aString = aString.substring(0, aString.lastIndexOf(";"));
        }
        // using Java 1.4 you can simply use split()
        // return aString.split("\\|");

        // to be compatible with Java 1.1 we split by hand
        Vector result = new Vector();
        int pos;
        while ((pos = aString.indexOf("|")) >= 0) {
            result.addElement(aString.substring(0,pos));
            aString = aString.substring(pos+1);
        }
        result.addElement(aString);
        return (String []) result.toArray(new String [result.size()]);
    }

    /**
     * Split GnuGk list of aliases (separated by equals char).
     * @param aString   string to split
     * @return          array of aliases
     */
    private static String [] splitAliases(String aString) {
        // using Java 1.4 you can simply use split()
        // return aString.split("=");

        // to be compatible with Java 1.1 we split by hand
        Vector result = new Vector();
        int pos;
        while ((pos = aString.indexOf("=")) >= 0) {
            result.addElement(aString.substring(0,pos));
            aString = aString.substring(pos+1);
        }
        result.addElement(aString);
        return (String []) result.toArray(new String [result.size()]);
    }

    /**
     * request current state from GnuGk.
     */
	private void refresh() {
		getCurrentRegistrations();
		getCurrentCalls();
	}

    public void getCurrentRegistrations() {
        sendGkCommand("PrintAllRegistrations");
    }

    public void getCurrentCalls() {
        sendGkCommand("PrintCurrentCalls");
    }

    /**
     * Send DisconnectEndpoint command to GnuGk.
     * @param _epid endpoint ID
     */
	public void disconnectEndpoint(String _epid) {
		sendGkCommand("DisconnectEndpoint " + _epid);
	}

    /**
     * Send DisconnectAlias command to GnuGk.
     * @param _alias endpoint alias
     */
	public void disconnectAlias(String _alias) {
	    sendGkCommand("DisconnectAlias " + H323Utils.extractAliasName(_alias));
	}
	
    /**
     * Send UnregisterAlias command to GnuGk.
     * @param _alias endpoint alias
     */
	public void unregisterAlias(String _alias) {
	    sendGkCommand("UnregisterAlias " + H323Utils.extractAliasName(_alias));
	}

    /**
     * Send MakeCall command to GnuGk.
     * @param _epid         endpoint ID
     * @param _destination  destination alias
     */
	public void makeCall(String _epid, String _destination) {
	    sendGkCommand("MakeCall " + _epid + " " + _destination);
	}

    /**
     * Send TransferCall command to GnuGk.
     * @param _alias        alias to stay in call
     * @param _destination  new destination to transfer the call to
     */
    public void transferCall(String _alias, String _destination) {
        sendGkCommand("TransferCall " + H323Utils.extractAliasName(_alias) + " " + H323Utils.extractAliasName(_destination));
    }

    /**
     * send RouteToAlias response to GnuGk.
     * @param _alias        endpoint alias
     * @param callingEpId   endpoint ID of calling endpoint
     * @param callRef       call reference
     */
    public void routeToAlias(String _alias, String callingEpId, String callRef) {
        sendGkCommand("RouteToAlias " + H323Utils.extractAliasName(_alias) + " " + callingEpId + " " + callRef);
    }

    /**
     * send RouteReject response to GnuGk.
     * @param callingEpId   endpoint ID of calling endpoint
     * @param callRef       call reference
     */
    public void routeReject(String callingEpId, String callRef) {
        sendGkCommand("RouteReject " + callingEpId + " " + callRef);
    }

    /**
     * helper function to send bytes to GnuGk.
     * @param cmd
     */
    private void sendGkCommand(String cmd) {
        if (ostream != null) {
            try {
                Logger.debug("ACD->GK:" + cmd);
                ostream.write((cmd + "\r\n").getBytes());
            }
            catch (IOException e) { Logger.log("Exception: " + e); }
        }
    }

}
