// This work is published under the GNU Public License (GPL)
// see file COPYING for details.
//
// The initial creation of this ACD application has been
// sponsored by IPTEL Inc. www.iptel.com.
//
// Author: Jan Willamowius <jan@willamowius.de>

package org.gnugk.acd;

import org.gnugk.GnuGkAppConfig;

import java.util.MissingResourceException;
import java.util.ResourceBundle;
import java.util.Enumeration;
import java.util.Vector;

/**
 * AgentConfig is a Singleton class to deal with the ACD configuration.
 *
 * @author Jan Willamowius <jan@willamowius.de>
 * @version 1.1
 */

public class ACDConfig extends GnuGkAppConfig {
    private static ACDConfig instance = new ACDConfig();
    private ResourceBundle agentConfig;

    // constants for distribution algorithms
    public final static int FIRST_FIT = 1;
    public final static int ROUND_ROBIN = 2;
    public final static int LONGEST_IDLE = 3;

    // constants for queueing modes
    public final static int QUEUEING_MODE_REJECT = 1;
    public final static int QUEUEING_MODE_RINGING = 2;
    public final static int QUEUEING_MODE_TALKING = 3;
    public final static int QUEUEING_MODE_RINGANDTALK = 4;

    /**
     * Singleton access method.
     * @return the object
     */
    public static ACDConfig getInstance() {
        return instance;
    }

    /**
     * constructor must be private to ensure Singleton property.
     */
    private ACDConfig() {
        // load config file
        readConfig("ACDConfig");
        try {
            agentConfig = ResourceBundle.getBundle("ACDAgentConfig");
        } catch (MissingResourceException e) {
            System.out.println("Can find config file: Place ACDAgentConfig.properties on CLASSPATH !");
            System.exit(-1);
        }
    }

    /**
     * return list of queues.
     * @return list of queues
     */
    public String[] getQueues() {
        Vector queues = new Vector();
        Enumeration keys = config.getKeys();
        while (keys.hasMoreElements()) {
            // TODO: more reliable list of queues (maybe parse agents)
            String key = (String)keys.nextElement();
            if (key.endsWith("_queueing_mode")) {
                String queue = key.substring(0, key.indexOf("_queueing_mode"));
                queues.add(queue);
            }
        }
        return (String[]) queues.toArray(new String [queues.size()]);
    }

    /**
     * Distribution algorithem selected in config file.
     * @return constant for algorithm
     */
    public int getDistributionAlgorithm() {
        String algo = getConfigKey("distribution_algorithm", "FIRST_FIT");
        if (algo.equalsIgnoreCase("ROUND_ROBIN"))
            return ROUND_ROBIN;
        else if (algo.equalsIgnoreCase("LONGEST_IDLE"))
            return LONGEST_IDLE;
        else
            return FIRST_FIT;
    }

    /**
     * Read queueing mode selected in config file for this queue.
     * @param queue queue name
     * @return constant for mode
     */
    public int getQueueingMode(String queue) {
        String mode = getConfigKey(queue + "_queueing_mode", "REJECT");
        if (mode.equalsIgnoreCase("RINGING"))
            return QUEUEING_MODE_RINGING;
        else if (mode.equalsIgnoreCase("TALKING"))
            return QUEUEING_MODE_TALKING;
        else if (mode.equalsIgnoreCase("RINGANDTALK"))
            return QUEUEING_MODE_RINGANDTALK;
        else
            return QUEUEING_MODE_REJECT;
    }

    /**
     * Get ring timeout for this queue.
     * @param queue queue name
     * @return  timeout in seconds
     */
    public long getRingTimeout(String queue) {
        int timeout = getConfigKey(queue + "_ring_timeout", 60);    // 1 min
        return timeout;
    }

    /**
     * Get talk timeout for this queue.
     * @param queue queue name
     * @return  timeout in seconds
     */
    public long getTalkTimeout(String queue) {
        int timeout = getConfigKey(queue + "_talk_timeout", 3600);  // 1 hr
        return timeout;
    }

    /**
     * Get priority setting for this queue.
     * @param queue queue name
     * @return  priority
     */
    public int getPriority(String queue) {
        int prio = getConfigKey(queue + "_priority", 0);
        return prio;
    }

    /**
     * Get max size for this queue.
     * @param queue queue name
     * @return  size
     */
    public int getMaxSize(String queue) {
        int size = getConfigKey(queue + "_size", -1);
        return size;
    }

    /**
     * Get the Agent configuration.
     * @return  the resource bundle
     */
    public ResourceBundle getAgentConfig() {
        return agentConfig;
    }

}
