// This work is published under the GNU Public License (GPL)
// see file COPYING for details.
//
// The initial creation of this ACD application has been
// sponsored by IPTEL Inc. www.iptel.com.
//
// Author: Jan Willamowius <jan@willamowius.de>

package org.gnugk.acd;

import org.gnugk.H323Utils;
import org.gnugk.Alias;

import java.util.ResourceBundle;
import java.util.MissingResourceException;
import java.util.Vector;

/**
 * This class contains additional ACD data about an endpoint.
 *
 * @author Jan Willamowius <jan@willamowius.de>
 * @version 1.0
 */

public class Agent extends Alias {
    private long lastCall;
    private boolean blocked = false;
    private static ResourceBundle agentConfig = null;

    /**
     * Create an Agent object.
     * @param _epid Endpoint ID
     * @param _alias H.323 alias
     */
    public Agent(String _epid, String _alias) {
        super(_alias, _epid);
        setLastCall();  // start counting free time from object creation
        if (agentConfig == null)
            // load agent only once, when creating the first agent
            agentConfig = ACDConfig.getInstance().getAgentConfig();
    }

    /**
     * Get the time since last call in milliseconds.
     * @return elapsed time
     */
    public long getLastCall() {
        return lastCall;
    }

    /**
     * Set a specific time for the last call.
     * @param lastCall the time (in milliseconds)
     */
    public void setLastCall(long lastCall) {
        this.lastCall = lastCall;
    }

    /**
     * Set lastCall property to now.
     */
    public void setLastCall() {
        setLastCall(System.currentTimeMillis());
    }

    public String toString() {
        return super.toString();
    }

    /**
     * Query if agent is blocked for ACD work.
     * @return T/F
     */
    public boolean isBlocked() {
        return blocked;
    }

    /**
     * block / unblock agent from ACD work.
     * @param blocked   T/F
     */
    public void setBlocked(boolean blocked) {
        this.blocked = blocked;
    }

    /**
     * isAgent checks if this is an agent that is mentioned in the ACD config.
     * This is usefull, because the agent list also includes all other endpoints
     * registered with the gatekeeper.
     * @return true / false
     */
    public boolean isAgent() {
        String configStr = "";
        try {
            configStr = agentConfig.getString(H323Utils.extractAliasName(getAlias()));
        } catch (MissingResourceException e) {
            // agent not in config
            return false;
        }
        return (configStr.length() > 0);
    }

    /**
     * isAvailable checks if the Agent is not busy (talking) and is suited for
     * this particular queue.
     * @param queue the queue
     * @return true / false
     */
    public boolean isAvailable(String queue) {
        String configStr = "";
        if ((getState() == AVAILABLE) && !isBlocked()) {
            // check if agent can work this queue
            try {
                configStr = agentConfig.getString(H323Utils.extractAliasName(getAlias()));
            } catch (MissingResourceException e) {
                // agent not in config -> not available
                return false;
            }
            String [] groups = split_groups(configStr);
            for (int i=0; i < groups.length; i++) {
                if (groups[i].equals(queue))
                    return true;
            }
        }
        return false;
    }

    /**
     * Split string entries at the comma and remove whitespace.
     * @param aString   string to split
     * @return          array of fields
     */
    private static String [] split_groups(String aString) {
        // using Java 1.4 you can simply use split()
        // return aString.split(",");

        // to be compatible with Java 1.1 we split by hand
        Vector result = new Vector();
        int pos;
        while ((pos = aString.indexOf(",")) >= 0) {
            result.addElement(aString.substring(0,pos).trim());
            aString = aString.substring(pos+1);
        }
        result.addElement(aString.trim());
        return (String []) result.toArray(new String [result.size()]);
    }


}
