// This work is published under the GNU Public License (GPL)
// see file COPYING for details.
//
// The initial creation of this ACD application has been
// sponsored by IPTEL Inc. www.iptel.com.
//
// Author: Jan Willamowius <jan@willamowius.de>

package org.gnugk.acd;

import org.gnugk.GkClient;
import org.gnugk.Logger;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.Iterator;

/**
 * This class stores calls that could not be routed, yet.
 *
 * @author Jan Willamowius <jan@willamowius.de>
 * @version 1.0
 */
class CallBacklog {
    private GkClient gkclient = null;
    private ACDConfig config = null;
    private SortedSet pendingList = new TreeSet();

    public CallBacklog(GkClient _gkclient) {
        gkclient = _gkclient;
        config = ACDConfig.getInstance();
    }

    /**
     * Sore a new pending request if there is still rome in the queue, otherwise reject it.
     * @param queue         called queue
     * @param callerEndId   caller endpoint ID
     * @param callRef       callRef
     * @param callerAlias   caller H.323 alias
     * @param callerIp      caller IP
     * @param status        PendingRequest.RINGING / TALKING
     */
    public synchronized void store(String queue, String callerEndId, String callRef, String callerAlias, String callerIp, int status) {
        int maxSize = config.getMaxSize(queue);
        if ((maxSize <= 0) || (pendingList.size() < maxSize)) {
            int priority = config.getPriority(queue);
            long timeout = System.currentTimeMillis();
            if (status == PendingRequest.TALKING)
                timeout += (config.getTalkTimeout(queue) * 1000);
            else
                timeout += (config.getRingTimeout(queue) * 1000);
            PendingRequest request = new PendingRequest(priority, queue, callerEndId, callRef, callerAlias, callerIp, timeout, status);
            Logger.log("Call from " + callerAlias + " to " + queue + " queued.");
            pendingList.add(request);
        } else {
            Logger.log("Queue for " + queue + " is full - rejecting call.");
            gkclient.routeReject(callerEndId, callRef);
        }
    }

    /**
     * Check if requests have expired and delete + reject them.
     */
    public synchronized void expire() {
        long now = System.currentTimeMillis();
        for (Iterator i = pendingList.iterator(); i.hasNext(); ) {
            PendingRequest r = (PendingRequest) i.next();
            if (r.timeout < now) {
                String wqName;
                if (r.status == PendingRequest.TALKING) {
                    Logger.log("Pending call from " + r.callerAlias + " to " + r.queue + " expired in talking queue.");
                    i.remove();
                    gkclient.disconnectAlias(r.callerAlias);
                } else if ( (config.getQueueingMode(r.queue) == ACDConfig.QUEUEING_MODE_RINGANDTALK) &&    // transfer if: queue mode is talk after ring
                     (r.status == PendingRequest.RINGING) &&                                        // AND the call is already ringing
                     ( ((wqName = GnuGkACD.getInstance().getWaitQueue(r.queue)) != null) ) ) {             // AND we have a talinmg queue available
                    Logger.log("Pending call from " + r.callerAlias + " to " + r.queue + " routed to WQ " + wqName + " after ring timeout.");
                    gkclient.routeToAlias(wqName, r.callerEndId, r.callRef);
                    r.status = PendingRequest.TALKING;
                    r.timeout = System.currentTimeMillis() + (config.getTalkTimeout(r.queue) * 1000);
                } else {
                    Logger.log("Pending call from " + r.callerAlias + " to " + r.queue + " expired.");
                    i.remove();
                    gkclient.routeReject(r.callerEndId, r.callRef);
                }
            }
        }
    }

   /**
     * Check if we have a pending call this agent could handle.
     * @param agent available Agent
     */
    public synchronized void checkPending(Agent agent) {
        for (Iterator i = pendingList.iterator(); i.hasNext(); ) {
            PendingRequest r = (PendingRequest) i.next();
            if (agent.isAvailable(r.queue)) {
                String alias = agent.getAlias();
                Logger.log("Pending call from " + r.callerAlias + " to " + r.queue + " routed to " + alias);
                i.remove();
                agent.setState(Agent.TALKING);
                agent.setLastCall();
                if (r.status == PendingRequest.RINGING)
                    gkclient.routeToAlias(alias, r.callerEndId, r.callRef);
                else
                    gkclient.transferCall(r.callerAlias, alias);
                return;
            }
        }
    }

    /**
     * Remove all pending requests for this endpoint ID.
     * @param epid  endpoint ID
     */
    public synchronized void removeAllRequests(String epid) {
        for (Iterator i = pendingList.iterator(); i.hasNext(); ) {
            PendingRequest r = (PendingRequest) i.next();
            if (epid.equals(r.callerEndId)) {
                i.remove();
            }
        }
    }

    /**
     * Dump the list of all pending requests for debug.
     * @return  formatted list
     */
    public synchronized String dumpPendingList() {
        String result = "Pending list (" + pendingList.size() + " entries)\n";
        for (Iterator i = pendingList.iterator(); i.hasNext(); ) {
            PendingRequest r = (PendingRequest) i.next();
            result += r.queue + " (" + r.priority + ") " + r.callerAlias + " (" + r.callerEndId + "/" +r.callRef + ")\n";
        }
        return result;
    }
}
