// This work is published under the GNU Public License (GPL)
// see file COPYING for details.
//
// The initial creation of this ACD application has been
// sponsored by IPTEL Inc. www.iptel.com.
//
// Author: Jan Willamowius <jan@willamowius.de>

package org.gnugk.acd;

import org.gnugk.*;
import java.util.Vector;
import java.util.Date;
import java.text.SimpleDateFormat;    // stay with Java 1.1 containers for now, so MS JView can run the app

/**
 * This is the main class for the ACD application.
 * It stores the agent data and implements the routing algorithms.
 *
 * @author Jan Willamowius <jan@willamowius.de>
 * @version 1.2
 */

public class GnuGkACD implements CallRouter {
    private static GnuGkACD instance;   // Singleton

    private Vector aliasList = new Vector();
	private GkClient gkclient;
    private CallBacklog backlog;
    private HouseKeeper houseKeeping = new HouseKeeper();
    private WQManager waitQueues = new WQManager();
    private ACDCommandListener commandListener;
    private ACDConfig config;
    private int distribution;   // call distribution algorithm
    private String gkHost;
    private int gkPort;
    private int listenPort;
    private String redirLocation;

    /**
     * Start the ACD.
     * @param args  unused
     */
	public static void main(String[] args) {
		if (args.length > 0) {
			System.out.println("usage: java org.gnugk.acd.GnuGkACD\n");
			System.exit(1);
		}		

		instance = new GnuGkACD();
	}

    /**
     * Singleton access.
     * @return singleton object
     */
    public static GnuGkACD getInstance() {
        return instance;
    }

    private GnuGkACD() {
        // read ACD config
        config = ACDConfig.getInstance();
        gkHost = config.getConfigKey("gnugk_host", "127.0.0.1");
        gkPort = config.getConfigKey("gnugk_port", 7000);
        listenPort = config.getConfigKey("listen_port", 7001);
        redirLocation = config.getConfigKey("redir_location", "http://www.google.com/");
        distribution = config.getDistributionAlgorithm();
        // connect to the gatekeeper
        gkclient = new GkClient();
		gkclient.init(this, gkHost, gkPort);
        backlog = new CallBacklog(gkclient);
        houseKeeping.init(this, backlog);
        String logfile = config.getConfigKey("logfile", "");
        int debugLevel = config.getConfigKey("debug", 0);
        Logger.setDebug(debugLevel);
        Logger.log("--- GnuGk ACD config ---");
        Logger.log("GnuGk host: " + gkHost);
        Logger.log("GnuGk status port: " + gkPort);
        Logger.log("ACD listen port: " + listenPort);
        Logger.log("ACD distribution algo: " + distribution);
        Logger.log("ACD redir location: " + redirLocation);
        Logger.log("ACD logfile: " + logfile);
        Logger.log("ACD debug level: " + debugLevel);
        Logger.log("--- starting ACD ---");
        if (logfile.length() > 0) {
            Logger.setLogfile(logfile);
        }
		new Thread(gkclient).start();
        waitQueues.startWaitQueues(gkclient);
        // start listening for commands to the ACD
        commandListener = new ACDCommandListener();
        commandListener.init(listenPort, redirLocation);
        new Thread(commandListener).start();
        new Thread(houseKeeping).start();
    }

	public void addAlias(String _alias, String _epid) {
		if (aliasById(_alias, _epid) == null) {
            Agent newAgent = new Agent(_epid, _alias);
			aliasList.addElement(newAgent);
            // check if we have pending calls for this now available agent
            backlog.checkPending(newAgent);
		}
	}

    /**
     * Endpoint has unregistered, get rid of all it's aliases.
     * Also cancel all pending calls.
     * @param _epid
     */
    public void removeAllAliases(String _epid) {
		for(int i = 0; i < aliasList.size(); i++ ) {
			Alias alias = (Alias) aliasList.elementAt(i);
			if (_epid.equals(alias.getEpid())) {
				aliasList.removeElementAt(i);
			}
		}
        backlog.removeAllRequests(_epid);
	}

    /**
     * Set _all_ aliases on this endpoint with this CRV to available.
     * @param _epid endpoint ID
     * @param _crv  CRV
     */
    public void setAliasAvailable(String _epid, String _crv) {
        Agent agent = agentByCrv(_epid, _crv);  // find first alias
        while ((agent != null) && (_crv.length() > 0)) {
            agent.setState(Alias.AVAILABLE);
            agent.setLastCall();
            agent.setCrv("");
            // check if we have pending calls for this now available agent
            backlog.checkPending(agent);
            agent = agentByCrv(_epid, _crv);    // find next alias
        }
    }

    /**
     * Set this alias on this endpoint to TALKING and remember the CRV.
     * @param _alias    H.323 alias
     * @param _crv      CRV
     */
    public void setAliasTalking(String _alias, String _crv) {
        Alias alias = aliasByName(_alias);
        if (alias != null) {
            alias.setState(Alias.TALKING);
            alias.setCrv(_crv);
        }
	}

    public void clearAllAliasStates() {
        aliasList.clear();
 	}

    /**
     * Do a cross check if the agent states are OK.
     */
    public synchronized void agentSanityCheck() {
        for(int i = 0; i < aliasList.size(); i++ ) {
            Alias alias = (Alias) aliasList.elementAt(i);
            alias.setState(Alias.AVAILABLE);
        }
        gkclient.getCurrentCalls();
    }

    private Alias aliasByName(String _alias) {
        for(int i = 0; i < aliasList.size(); i++ ) {
            Alias alias = (Alias) aliasList.elementAt(i);
            if (_alias.equals(alias.getAlias())) {
                return alias;
            }
        }
        return null;
    }

    private Alias aliasById(String _alias, String _epid) {
        for(int i = 0; i < aliasList.size(); i++ ) {
            Alias alias = (Alias) aliasList.elementAt(i);
            if (_alias.equals(alias.getAlias()) && _epid.equals(alias.getEpid())) {
                return alias;
            }
        }
        return null;
    }

    private Agent agentByCrv(String _epid, String _crv) {
        for(int i = 0; i < aliasList.size(); i++ ) {
            Agent agent = (Agent) aliasList.elementAt(i);
            if (_epid.equals(agent.getEpid()) && _crv.equals(agent.getCrv()) ) {
                return agent;
            }
        }
        return null;
    }
	
    public void routeCall(String queue, String callerEndId,
                              String callRef, String callerAlias, String callerIp) {

        Agent agent = null;
        if (distribution == ACDConfig.FIRST_FIT) {
            agent = firstFitRouter(queue);
        } else if (distribution == ACDConfig.ROUND_ROBIN) {
            agent = roundRobinRouter(queue);
        } else if (distribution == ACDConfig.LONGEST_IDLE) {
            agent = longestIdleRouter(queue);
        }
        if (agent != null) {
            agent.setState(Agent.TALKING);
            agent.setLastCall();
            Logger.log("Routing call for " + queue + " to " + agent);
            gkclient.routeToAlias(agent.getAlias(), callerEndId, callRef);
        } else {
            int queueingMode = config.getQueueingMode(queue);
            String wqName;
            Logger.log("No agent available for call to " + queue + " (" + queueingMode + ")");
            if ( (queueingMode == ACDConfig.QUEUEING_MODE_RINGING) ||
                 (queueingMode == ACDConfig.QUEUEING_MODE_RINGANDTALK) ) {
                // store the request in the backlog, in case an agent becomes available before timeout
                backlog.store(queue, callerEndId, callRef, callerAlias, callerIp, PendingRequest.RINGING);
            } else if ( (queueingMode == ACDConfig.QUEUEING_MODE_TALKING) &&
                        ((wqName = getWaitQueue(queue)) != null) ) {
                gkclient.routeToAlias(wqName, callerEndId, callRef);
                // store the request in the backlog, in case an agent becomes available before timeout
                backlog.store(queue, callerEndId, callRef, callerAlias, callerIp, PendingRequest.TALKING);
            } else {
                // reject the call right away
                gkclient.routeReject(callerEndId, callRef);
            }
        }
    }

    /**
     * Find available wait queue for this queue.
     * @param queue name of called queue
     * @return  alias of wait queue agent or null if noe is found
     */
    public String getWaitQueue(String queue) {
        for(int i = 0; i < aliasList.size(); i++ ) {
            Agent agent = (Agent) aliasList.elementAt(i);
            if ( (agent.getAlias().startsWith(queue + "_wq_")) &&
                 (agent.getState() == Endpoint.AVAILABLE) ) {
                return agent.getAlias();
            }
        }
        return null;
    }

    // Algo: first-fit (unfair)
    private Agent firstFitRouter(String queue) {
        return firstFitRouter(0, queue);
    }

    // Algo: first-fit (unfair)
    private Agent firstFitRouter(int start, String queue) {
        if (start > aliasList.size()) {
            start = 0;
        }
        for(int i = start; i < aliasList.size(); i++ ) {
            Agent agent = (Agent) aliasList.elementAt(i);
            if (agent.isAvailable(queue)) {
                return agent;
            }
        }
        return null;
    }

    // Algo: round-robin (fair)
    private Agent lastPick = null;
    private Agent roundRobinRouter(String queue) {
        Agent agent = null;
        if (lastPick != null) {
            // find the agent we picked last
            for(int i = 0; i < aliasList.size(); i++ ) {
                if (lastPick.equals( aliasList.elementAt(i)) ) {
                    // find the next available agent (position to list end)
                    agent = firstFitRouter(i + 1, queue);
                    break;
                }
            }
        }
        if (agent == null) {
            // go around to the start and look again
            agent = firstFitRouter(0, queue);
        }
        lastPick = agent;
        return agent;
    }

    // Algo: longest-idle (fair)
    private Agent longestIdleRouter(String queue) {
        // pick the first available agent
        Agent longestIdle = firstFitRouter(0, queue);
        // see if we can find anyone that has been idle longer
        if (longestIdle != null) {  // no need to look if everybody is talking
            for(int i = 0; i < aliasList.size(); i++ ) {
                Agent agent = (Agent) aliasList.elementAt(i);
                if (agent.isAvailable(queue) &&
                   (agent.getLastCall() < longestIdle.getLastCall())) {
                    longestIdle = agent;
                }
            }
        }
        return longestIdle;
    }

    /**
     * agent is not available for ACD work.
     * @param agent
     */
    public void logout(String agent) {
        for(int i = 0; i < aliasList.size(); i++ ) {
            Agent ag = (Agent) aliasList.elementAt(i);
            if (agent.equals(H323Utils.extractAliasName(ag.getAlias()))) {
                ag.setBlocked(true);
                Logger.log("Logout agent " + agent);
            }
        }
    }

    /**
     * set agent to be available for ACD work again.
     * @param agent
     */
    public void login(String agent) {
        for(int i = 0; i < aliasList.size(); i++ ) {
            Agent ag = (Agent) aliasList.elementAt(i);
            if (agent.equals(H323Utils.extractAliasName(ag.getAlias()))) {
                ag.setBlocked(false);
                Logger.log("Login agent " + agent);
            }
        }
    }

    /**
     * Dump the list of agents with their current state.
     * @return  formatted list
     */
    public String dumpAgentStates() {
        SimpleDateFormat formatter = new SimpleDateFormat("HH:mm:ss");
        String result = "Agent list\n";
        for(int i = 0; i < aliasList.size(); i++ ) {
            Agent ag = (Agent) aliasList.elementAt(i);
            if (ag.isAgent()) {
                // agent list contains all GK registration, only list configured agents
                result += ag.getAlias() + " (" + ag.getEpid() + "):\t" +
                        (ag.isBlocked() ? "NOT logged in" : "    logged in") +
                        "\tstate: " + (ag.getState()==0 ? "AVAILABLE" : "TALKING   " ) +
                        "\tlast call: " + formatter.format(new Date(ag.getLastCall())) +
                        "\n";
            }
        }
        return result;
    }

    /**
     * Dumnp list of pending calls.
     * @return  formatted list
     */
    public String dumpPendingList() {
        return backlog.dumpPendingList();
    }

}
